function [J, mm_bar, LL_bar, Lsigma_w_bar] = svi_grad(mm, LL, Lsigma_w, neg_log_like_grad)
%SVI_GRAD cost function and gradients for black-box stochastic variational inference
%
%     [J, mm_bar, LL_bar, Lsigma_w_bar] = svi_grad(mm, LL, Lsigma_w, neg_log_like_grad)
%
% Inputs:
%                    mm Dx1 mean of variational posterior
%                    LL DxD lower-triangular Cholesky decomposition of
%                           variational posterior, with diagonal log-transformed
%              Lsigma_w 1x1 log of prior standard deviation over weights
%     neg_log_like_grad @fn -ve log-likelihood of model and gradients wrt weights
%                           Could be an unbiased estimate based on a mini-batch,
%                           we only get unbiased estimates of cost and gradients anyway.
%
% Outputs:
%                     J 1x1 estimate of variational cost function = -ELBO
%                mm_bar Dx1 with derivatives wrt mm, ...
%                LL_bar DxD ...LL, ...
%          Lsigma_w_bar 1x1 ...and Lsigma_w

% Iain Murray, November 2016

% Unpack Cholesky factor of posterior covariance and prior variance
% from their unconstrained forms.
D = numel(mm);
L = tril(LL); L(1:(D+1):end) = exp(LL(1:(D+1):end));
sigma2_w = exp(2*Lsigma_w);

% The estimate of the variational cost function
J1 = -0.5*D - sum(diag(LL)); % - D/2*log(2*pi)
tmp = (L(:)'*L(:) + mm'*mm)/sigma2_w;
J2 = tmp/2 + D*Lsigma_w;   % + D/2*log(2*pi)
nu = randn(D, 1);
ww = mm + L*nu; % Using random weight
[J3, ww_bar] = neg_log_like_grad(ww);
J = J1 + J2 + J3;

% The derivatives
mm_bar = mm/sigma2_w + ww_bar;
L_bar = L/sigma2_w + tril(ww_bar*nu');
LL_bar = L_bar;
LL_bar(1:(D+1):end) = L_bar(1:(D+1):end) .* L(1:(D+1):end) - 1;
Lsigma_w_bar = D - tmp;

